import pytest
import uuid

from calitp.config import get_bucket
from calitp.storage import get_fs

GCS_BUCKET = "gs://calitp-py-ci"


@pytest.fixture
def tmp_name():
    # generate a random table name. ensure it does not start with a number.
    name = f"{GCS_BUCKET}/calitp-" + str(uuid.uuid4())
    yield name

    try:
        fs = get_fs()
        fs.rm(tmp_name, recursive=True)
    except Exception:
        UserWarning("GCS folder not deleted: %s" % name)


def test_get_fs_pipe(tmp_name):
    bucket = get_bucket()
    bucket_dir = f"{bucket}/calitp/{tmp_name}"
    fname = f"{bucket_dir}/test.txt"

    fs = get_fs()
    fs.pipe(fname, "1".encode())

    res = fs.listdir(bucket_dir)
    assert len(res) == 1

    fs.cat(fname) == "1"



import pandas as pd
import pytest
import uuid

from calitp.sql import get_table, write_table, query_sql
from calitp.config import RequiresAdminWarning, pipeline_context
from calitp.tests.helpers import CI_SCHEMA_NAME

from pandas.testing import assert_frame_equal

# TODO: set up a separate project for CI, so our CI doesn't have permission
# to create / delete prod tables. Bigquery lets you set read access on individual
# datasets, but write access happens at the project level. Working around now
# by creating a table outside CI, then only testing reading it.


@pytest.fixture
def tmp_name():
    from sqlalchemy.exc import NoSuchTableError

    # generate a random table name. ensure it does not start with a number.
    table_name = "t_" + str(uuid.uuid4()).replace("-", "_")
    schema_table = f"{CI_SCHEMA_NAME}.{table_name}"

    yield schema_table

    try:
        tbl = get_table(schema_table)
        tbl.drop()
    except NoSuchTableError:
        pass


@pytest.mark.skip
def test_write_table(tmp_name):
    df = pd.DataFrame({"x": [1, 2, 3]})

    with pipeline_context():
        write_table(df, tmp_name)


def test_write_table_no_admin(tmp_name):
    df = pd.DataFrame({"x": [1, 2, 3]})

    with pytest.warns(RequiresAdminWarning):
        write_table(df, tmp_name)


def test_get_table(tmp_name):
    df = pd.DataFrame({"x": [1, 2, 3]})
    with pipeline_context():
        write_table(df, tmp_name)

    tbl_test = get_table(tmp_name, as_df=True)
    assert_frame_equal(tbl_test, df)


def test_query_sql():
    import tempfile
    import datetime

    from pathlib import Path
    from calitp.templates import user_defined_macros

    with tempfile.TemporaryDirectory() as dir_name:
        p_query = Path(dir_name) / "query.yml"
        p_query.write_text("""sql: SELECT {{THE_FUTURE}}""")

        query_txt = query_sql(str(p_query), dry_run=True, as_df=False)
        assert query_txt == f"SELECT {user_defined_macros['THE_FUTURE']}"

        res = query_sql(str(p_query), as_df=False)
        assert res.fetchall() == [(datetime.date(2099, 1, 1),)]


def test_query_sql_as_df():
    df = query_sql("SELECT 1 AS n")
    assert len(df) == 1
    assert "n" in df.columns



import os
import warnings

from contextlib import contextmanager
from pathlib import Path
from functools import wraps


CALITP_BQ_MAX_BYTES = os.environ.get("CALITP_BQ_MAX_BYTES", 5_000_000_000)
CALITP_BQ_LOCATION = os.environ.get("CALITP_BQ_LOCATION", "us-west2")


class RequiresAdminWarning(UserWarning):
    """Represents errors where, e.g., a user tries to load a table to the warehouse."""


def is_development():
    options = {"development", "cal-itp-data-infra"}

    # if a person can write data, then they need to set AIRFLOW_ENV
    if is_pipeline():
        if "AIRFLOW_ENV" not in os.environ:
            raise KeyError(
                "Pipeline admin must set AIRFLOW_ENV env variable explicitly"
            )

        env = os.environ["AIRFLOW_ENV"]

        if env not in options:
            raise ValueError("AIRFLOW_ENV variable must be one of %s" % options)

    # otherwise, analysts connect to prod data by default
    else:
        env = os.environ.get("AIRFLOW_ENV", "cal-itp-data-infra")

    return env == "development"


def is_cloud():
    return os.environ.get("CALITP_AUTH") == "cloud"


def get_bucket():
    # TODO: can probably pull some of these behaviors into a config class
    if is_development():
        return "gs://gtfs-data-test"
    else:
        return "gs://gtfs-data"


def get_project_id():
    if is_development():
        return "cal-itp-data-infra-staging"

    return "cal-itp-data-infra"


def is_pipeline():
    return os.environ.get("CALITP_USER") == "pipeline"


def require_pipeline(func_name):
    """Decorator that skips a function is user is not admin.

    Note: this function is for convenience (not to replace appropriate access levels)
    """

    if not isinstance(func_name, str):
        raise TypeError("Warning must be a string")

    warning = f"Not running in pipeline, so skipping {func_name}()"

    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            if not is_pipeline():
                warnings.warn(warning, RequiresAdminWarning)
            else:
                return f(*args, **kwargs)

        return wrapper

    return decorator


def format_table_name(name, is_staging=False, full_name=False):
    dataset, table_name = name.split(".")
    staging = "__staging" if is_staging else ""
    # test_prefix = "zzz_test_" if is_development() else ""
    test_prefix = ""

    project_id = get_project_id() + "." if full_name else ""
    # e.g. test_gtfs_schedule__staging.agency

    return f"{project_id}{test_prefix}{dataset}.{table_name}{staging}"


def pipe_file_name(path):
    """Returns absolute path for a file in the pipeline (e.g. the data folder).

    """

    # For now, we just get the path relative to the directory holding the
    # DAGs folder. For some reason, gcp doesn't expose the same variable
    # for this folder, so need to handle differently on dev.
    if is_development():
        root = Path(os.environ["AIRFLOW__CORE__DAGS_FOLDER"]).parent
    else:
        root = Path(os.environ["DAGS_FOLDER"]).parent

    return str(root / path)


@contextmanager
def pipeline_context():
    """Temporarily set CALITP_USER to be pipeline.

    This allows a user to write to tables on production and staging.
    """

    prev_user = os.environ.get("CALITP_USER")

    os.environ["CALITP_USER"] = "pipeline"

    try:
        yield
    finally:
        if prev_user is None:
            del os.environ["CALITP_USER"]
        else:
            os.environ["CALITP_USER"] = prev_user



import pytest
import uuid
import pandas as pd

from calitp.sql import get_table, write_table, get_engine
from calitp.tables import AutoTable
from siuba.sql import LazyTbl

from contextlib import contextmanager
from calitp.tests.helpers import CI_SCHEMA_NAME


# TODO: copied from tests.sql, consolidate into helpers
@contextmanager
def as_pipeline():
    import os

    prev_user = os.environ.get("CALITP_USER")

    os.environ["CALITP_USER"] = "pipeline"

    try:
        yield
    finally:
        if prev_user is None:
            del os.environ["CALITP_USER"]
        else:
            os.environ["CALITP_USER"] = prev_user


@pytest.fixture
def tmp_name():
    from sqlalchemy.exc import NoSuchTableError

    # Code that runs before test ----

    # generate a random table name. ensure it does not start with a number.
    table_name = "t_" + str(uuid.uuid4()).replace("-", "_")
    schema_table = f"{CI_SCHEMA_NAME}.{table_name}"

    # pass the name of the temporary table into the test
    yield schema_table

    # Code that runs after test ----
    # delete table corresponding to temporary name, if it exists
    try:
        tbl = get_table(schema_table)
        tbl.drop()
    except NoSuchTableError:
        pass


def test_write_table(tmp_name):
    schema_name, table_name = tmp_name.split(".")

    df = pd.DataFrame({"x": [1, 2, 3]})

    with as_pipeline():
        write_table(df, tmp_name)

    tbl = AutoTable(get_engine(), lambda s: s, lambda s: True,)  # s.replace(".", "_"),

    tbl._init()

    tbl_tmp = getattr(tbl.calitp_py, table_name)()

    assert isinstance(tbl_tmp, LazyTbl)


def test_auto_table_comments(tmp_name):
    from calitp.tables import TableFactory

    get_engine().execute(
        f"""
        CREATE TABLE `{tmp_name}` (
            x INT64 OPTIONS(description="x column"),
            y STRING OPTIONS(description="y column")
        )
        OPTIONS(
            description="the table comment"
        )
    """
    )

    # TODO: rather than using AutoTable, let's just use CalitpTable directly
    tbl_factory_tmp = TableFactory(get_engine(), tmp_name)
    doc = tbl_factory_tmp._repr_html_()

    assert "x column" in doc
    assert "the table comment" in doc
